#! /usr/bin//perl -w
#
# Copyright 1999  August Hoerandl (hoerandl@elina.htlw1.ac.at)
#                 http://elina.htlw1.ac.at/~hoerandl
#
# parts from search.pl which is
# Copyright 1999 Jose M. Vidal
# Jose M. Vidal, vidal@multiagent.com, http://jmvidal.ece.sc.edu
#
# This program is free software.  You can redistribute it and/or modify
# it under the terms of the GNU General Public License
#
# create database with meta info 
#
# actions
#  read database
#    read to @allRecords
#  read metadata
#    read to $meta{$URL}, $date{$URL}, $error{$URL} 
#  check all (expired) urls
#    based on $date{$URL}
#  dump metadata 
#  dump database with meta info
#  dump errordata

#$database   = "/usr/local/httpd/cgi-bin/urls.db";      # Name of db file.
$database = "urls.db";
$newdb = $database;      # result: database
$metadata = "meta.db";   # result: metainfo
$errordata = "error.db"; # result: urls with error

$error_recheck = 10; # number of days for recheck after error
$age_recheck = 30;  # number of days for recheck 
# add some randomness - so we won't fetch all urls on the same day 
$random = .5;   # random number of days added (factor)

$dumpfreq = 10;  # when to dump metadata (number of urls read) 
$maxurls = 2;   # max url to fetch recursive (frames)
$maxcheck = 1000; # max number of urls to check in one run

$sleep = 2; # time to sleep between url requests
$debug = 0; # set to 1 for debug output

# ------------------------------------------------------------------
# no user serviceable parts below
# ------------------------------------------------------------------

use LWP;
use HTTP::Request::Common;

# ------------------------------------------------------------------

sub read_database {

  open(DB,"<$database") ||
    die ("Can't open $database; $!");
  $db_line1 = <DB>;
  chop($db_line1);
  if ($db_line1 eq "#bk2site urls.db generated file"){
    $numlines = 9;
  }
  else {
    $numlines = 10;
  };
  $db_line2 = <DB>;
  $db_line3 = <DB>;
  $db_line4 = <DB>;
  @allRecords = <DB>; # there are numlines lines per record,
                      # so mutiply index by 8

  close(DB);
  ## 0 is the type
  ## 1 is the ParentTitle
  ## 2 is the url or relative dir
  ## 3 is the Title
  ## 4 is the comment
  ## 5 is the creation time
  ## 6 is the last modified time
  ## 7 is the last visit time
  ## 8 is the number of user hits
  ## 9 is the meta info

  $typeN = 0;
  #  $parentTitleN = 1;
  $urlN = 2;
  #  $titleN = 3;
  #  $commentN = 4;
  #  $creationtimeN = 5;
  #  $modifiedtimeN = 6;
  #  $visittimeN = 7;
  #  $urlhitsN = 8;
  
#  $numlines = 10; ##0-9, as above

  #  my $numRecords = 15;  # test only
  my $numRecords = ($#allRecords + 1)/$numlines;

  my $i;
  for ($i=0; $i < $numRecords; ++$i){ #initialize meta
    my $cr = ($i * $numlines);
    next if (($allRecords[$cr + $typeN]) eq "FOLDER\n");
    my $url =    $allRecords[$cr + $urlN];
    chop($url);
    print "url $url added\n" if $debug;
    $error{$url} = 0;
    $date{$url} = 0;
    $meta{$url} = "";
  }
}

sub dump_database {
  print "\ndumping database\n";
  open(DB,">$newdb") ||
    die ("Can't open $newdb; $!");
#  print DB "#bk2site urls.db generated file metadata";
  if ($db_line1 eq "#bk2site urls.db generated file") {
    print DB "#bk2site urls.db generated file metadata";
  }
  else {
    print DB $db_line1;
  }
  print DB "$db_line2";
  print DB "$db_line3";
  print DB "$db_line4";
  my $numRecords = ($#allRecords + 1)/$numlines;
  for ($i=0; $i < $numRecords; $i++) {
    my $cr = ($i * $numlines);
    my $url = $allRecords[$cr + $urlN];
    chop($url);
    for($j = 0; $j <  $numlines-1; $j++) {
      print DB $allRecords[$cr + $j];
    }
    if (defined $error{$url} && $error{$url} > 0) {
      print DB "\n";
    } elsif (defined $meta{$url}) {
      print DB "$meta{$url}\n";
    } else {
      print DB "\n";
    }
  }
  close DB;
}

# ------------------------------------------------------------------
# add url + filename
sub make_url {
  local($url, $file) = @_;
  if ( $file =~ /^\//) {  # filename starting with /
    $url =~ /(.*\/\/[^\/]*)\//; # get only host part
    return $1 . $file;
  }
  if ( $url =~ /\/$/ ) { # url ends with /
    return $url . $file;
  } else {
    return $url . "/" . $file;
  }
}


sub get_meta {
  local($url,$count) = @_;

  print "getting $url\n" if $debug;
  sleep($sleep) if ($sleep > 0);

  my $ua = LWP::UserAgent->new;
  $ua->env_proxy();
  my $page = $ua->request(GET "$url");
  my $metainfo = "";
  
  $error = "";
  if ($page->is_error) {
    $error = $page->status_line;
    return "";
  }

  $refresh = $page->freshness_lifetime;
  my $old = "";
  my @lines = split(/^/,$page->content());
  for (@lines) {
    last if /body/i;
    $_ = $old . $_;
    if (/meta/i && (/keywords/i || /description/i)) {
      if (/content *= *[\"\']([^\"\']*)[\"\']/i) {
	print "meta = $1\n" if $debug;
	$metainfo .= "$1 ";
      } else {
	print "oops (no content): $_\n" if $debug;
	# line with meta but no content - keep & merge with next line
	$old = chop($_);
      }
    } else {
      $old = "";
    }
    # look for: frame src="..."
    if (/frame/i && /src *= *\"([^\"]*)\"/i && $count < $maxurls) {
      my $tmp = get_meta( &make_url($page->base,$1), $count+1);
      if ($error eq "") {
	$metainfo .= $tmp;
      }
      $error = "";
    }
  }
  print "done $url\n" if $debug;
  return $metainfo;
}

sub check {
  local($url) = @_;
  $error = "";
  $res = &get_meta ($url, 1);
  if ($error ne "") { # this is a "bad" url
    print "  OOPS $error\n";
    $date{$url} = time + 
      24*60*60*($error_recheck + rand($random*$error_recheck));
    $error{$url}++;
    $meta{$url} = $error;
  } else { # we have found the url
    print " [$res]\n";
    $error{$url} = 0;
    my $w = time +  24*60*60*($age_recheck + rand($random*$age_recheck));
    if ($w > $refresh) {
      $date{$url} = $w;
    } else {
      $date{$url} = $refresh;
    }
    #  print "recheck in ", ($date{$url} - time)/24/60/60, " days\n";
    $meta{$url} = $res;
  }
}

# ------------------------------------------------------------------

sub read_metadata {
  open(DB,"<$metadata") ||
    return;
  print "reading metadata\n";
  while ($url = <DB>) {
    chop $url;
    $error= <DB>;
    $date = <DB>;
    $meta = <DB>;
    if (defined $meta{$url}) {
      chop $error;
      chop $date;
      chop $meta;
      $error{$url}=$error;
      $date{$url} =$date;
      $meta{$url} =$meta;
      print "added [$url]\n" if $debug;
    } else {
      print "skipped $url\n";
    } 
  }
  close(DB);
}


sub dump_metadata {
  open(DB,">$metadata") ||
    die ("Can't open $metadata; $!");
  print "\ndumping metadata\n";
  foreach $url (sort keys (%meta)) {
    print DB "$url\n";
    print DB "$error{$url}\n";
    print DB "$date{$url}\n";
    print DB "$meta{$url}\n";
    print "dumped [$url]\n" if $debug;
  }
  close(DB);
}

sub dump_errordata {
  open(DB,">$errordata") ||
    die ("Can't open $errordata; $!");
  print "\ndumping errordata\n";
  foreach $url (sort keys (%meta)) {
    if ($error{$url} > 0) {
      print DB "$url\n";
      print DB "$error{$url}\n";
      print DB "$meta{$url}\n\n";
      print "dumped [$url]\n" if $debug;
    }
  }
  close(DB);
}

# ------------------------------------------------------------------

sub check_all {
  print "\n\n";
  my $all = 0;
  foreach $url (keys (%meta)) {
    $all++;
  }
  my $i = 0;
  my $checked = 0;
  foreach $url (keys (%meta)) {
    $i ++;
    print "checking [$url]  ", $checked, "/", $i, "/", $all, "\n";
    if (time > $date{$url}) {
      &check($url);
      $checked ++;
      if ($checked % $dumpfreq == 0) {
	&dump_metadata;
	&dump_errordata;
      }
    } else {
      printf("  recheck in %.0f days\n",
	     ($date{$url}-time)/24/60/60);
    }
    last if ($checked > $maxcheck);
  }
}

# ------------------------------------------------------------------

sub testoneurl {
  # test only
  #check("http://localhost/~hoerandl/xxx.html");
  #check("http://localhost/~hoerandl/xxxx.html");
  #check("http://localhost/~hoerandl");
  #check("http://localhost/~hoerandl/tk");
  #check("http://elina/");
  #check("http://elina/start");
  #check("http://elina.htlw1.ac.at/linux");
  #check("http://www.linux.org/");
  #check("http://www.linux.de");
}

# ------------------------------------------------------------------
#
# program starts here
#

&read_database;
&read_metadata;

&check_all;

&dump_metadata;
&dump_database;
&dump_errordata;
